From 5fc8eaf271ae7019880fbb53b7ba51760b0ef596 Mon Sep 17 00:00:00 2001
From: Matthias van de Meent <boekewurm+postgres@gmail.com>
Date: Fri, 22 Nov 2024 17:50:15 +0100
Subject: [PATCH v0 2/2] Modernize bulk_write's internals with vectorized
 writes

By merging operations we reduce the number of syscalls and thus context
switches.  We now also make use of zeroextend, rather than our own artisinal
version using smgrextend() with an empty page; further increasing efficiency
in the storage layer.
---
 src/backend/storage/smgr/bulk_write.c | 79 ++++++++++++++++++---------
 1 file changed, 53 insertions(+), 26 deletions(-)

diff --git a/src/backend/storage/smgr/bulk_write.c b/src/backend/storage/smgr/bulk_write.c
index 274051c40d..1c40801e2b 100644
--- a/src/backend/storage/smgr/bulk_write.c
+++ b/src/backend/storage/smgr/bulk_write.c
@@ -46,8 +46,6 @@
 
 #define MAX_PENDING_WRITES XLR_MAX_BLOCK_ID
 
-static const PGIOAlignedBlock zero_buffer = {{0}};	/* worth BLCKSZ */
-
 typedef struct PendingWrite
 {
 	BulkWriteBuffer buf;
@@ -274,40 +272,69 @@ smgr_bulk_flush(BulkWriteState *bulkstate)
 					 npending, blknos, pages, page_std);
 	}
 
-	for (int i = 0; i < npending; i++)
+	while (npending > 0)
 	{
-		BlockNumber blkno = pending_writes[i].blkno;
-		Page		page = pending_writes[i].buf->data;
+		const void  *buffers[MAX_PENDING_WRITES];
+		int		nbufs = 0;
+
+		BlockNumber	blkno = pending_writes[nbufs].blkno;
+		buffers[nbufs] = pending_writes[nbufs].buf->data;
 
-		PageSetChecksumInplace(page, blkno);
+		PageSetChecksumInplace(pending_writes[nbufs].buf->data, blkno + nbufs);
+		nbufs++;
 
-		if (blkno >= bulkstate->relsize)
+		for (; nbufs < npending && nbufs < MAX_PENDING_WRITES;)
 		{
+			if (pending_writes[nbufs].blkno != blkno + nbufs)
+				break;
+
 			/*
-			 * If we have to write pages nonsequentially, fill in the space
-			 * with zeroes until we come back and overwrite.  This is not
-			 * logically necessary on standard Unix filesystems (unwritten
-			 * space will read as zeroes anyway), but it should help to avoid
-			 * fragmentation.  The dummy pages aren't WAL-logged though.
+			 * Never combine past the end of the relation; we need to separate
+			 * smgrwritev() and smgrextendv()
 			 */
-			while (blkno > bulkstate->relsize)
-			{
-				/* don't set checksum for all-zero page */
-				smgrextend(bulkstate->smgr, bulkstate->forknum,
-						   bulkstate->relsize,
-						   &zero_buffer,
-						   true);
-				bulkstate->relsize++;
-			}
-
-			smgrextend(bulkstate->smgr, bulkstate->forknum, blkno, page, true);
-			bulkstate->relsize++;
+			if (blkno + nbufs == bulkstate->relsize)
+				break;
+
+			buffers[nbufs] = pending_writes[nbufs].buf->data;
+			PageSetChecksumInplace(pending_writes[nbufs].buf->data,
+								   blkno + nbufs);
+			nbufs++;
+		}
+		
+		Assert(((int64) blkno) + nbufs <= InvalidBlockNumber);
+
+		if (blkno < bulkstate->relsize)
+		{
+			Assert(blkno + nbufs <= bulkstate->relsize);
+
+			smgrwritev(bulkstate->smgr, bulkstate->forknum, blkno,
+					   buffers, nbufs, true);
 		}
 		else
-			smgrwrite(bulkstate->smgr, bulkstate->forknum, blkno, page, true);
-		pfree(page);
+		{
+			Assert(blkno >= bulkstate->relsize);
+
+			if (blkno > bulkstate->relsize)
+				smgrzeroextend(bulkstate->smgr, bulkstate->forknum,
+							   bulkstate->relsize,
+							   (int) (blkno - bulkstate->relsize),
+							   true);
+
+			smgrextendv(bulkstate->smgr, bulkstate->forknum, blkno,
+						buffers, nbufs, true);
+
+			bulkstate->relsize = blkno + nbufs;
+		}
+
+		npending -= nbufs;
+		pending_writes += nbufs;
+
+		for (int i = 0; i < nbufs; i++)
+			pfree((void *) buffers[i]);
 	}
 
+	Assert(npending == 0);
+
 	bulkstate->npending = 0;
 }
 
-- 
2.45.2

